'' =================================================================================================
''
''   File....... jm_i2c_driven.spin
''   Purpose.... Basic I2C driver
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2009-2010 Jon McPhalen
''               -- several portions inspired by code from Mike Green
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 28 JUL 2009
''   Updated.... 04 MAY 2010
''
'' =================================================================================================

{{

   Low/Mid-level I2C routines.
 
   Assumes that SCL and SDA are fixed in a given application and can be set on
   initialization of the object.  Code drives SCL pin for compatibility with
   Parallax boards that do not have an SCL pullup.

   Use "x" routines for devices with 16-bit addresses
  
}}   


CON

   #0, ACK, NAK

  #28, BOOTSCL, BOOTSDA


VAR

  long  scl                                                     ' clock pin of i2c buss
  long  sda                                                     ' data pin of i2c buss
           

PUB init(sclpin, sdapin)

  scl := sclpin
  sda := sdapin


PUB putbyte(id, addr, value) | ackbit

'' Write byte to device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  wait(id)
  write(addr)                                                   ' write address
  ackbit := write(value)                                        ' write one byte
  stop

  return ackbit


PUB putword(id, addr, value) | ackbit

'' Write word to device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device  

  ackbit := putpage(id, addr, 2, @value)

  return ackbit


pub putlong(id, addr, value) | ackbit

'' Write long to device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  ackbit := putpage(id, addr, 4, @value)

  return ackbit
   

PUB putpage(id, addr, n, src) | ackbit

'' Write n bytes from src to device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device
'' --    n : bytes to write
'' --  src : pointer to source valuee
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr)                                                   ' lsb of address
  ackbit := ACK                                                 ' assume okay
  repeat n
    ackbit |= write(byte[src++])                                ' write a byte 
  stop

  return ackbit


PUB getbyte(id, addr) | value

'' Read byte from device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  wait(id)
  write(addr)                                                   ' write address
  start                                                         ' restart for read
  write(id | $01)                                               ' read  
  value := read(NAK)                                            '  one byte
  stop

  return value & $FF


PUB getword(id, addr) | value

'' Read word from device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  getpage(id, addr, 2, @value)

  return value & $FFFF


PUB getlong(id, addr) | value

'' Read long from device w/8-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  getpage(id, addr, 4, @value)

  return value 


PUB getpage(id, addr, n, dest)

'' Read n bytes from device w/8-bit addressing; output to dest
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
'' --    n : bytes to read
'' -- dest : pointer to the destination
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr)                                                   ' lsb of address
  start                                                         ' restart for read
  write(id | $01)                                               ' device read
  repeat (n - 1)
    byte[dest++] := read(ACK)
  byte[dest] := read(NAK)                                       ' last byte gets NAK
  stop 


PUB putbytex(id, addr, value) | ackbit

'' Write byte to device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 16-bit address in the device

  wait(id)
  write(addr.byte[1])                                           ' msb of address
  write(addr.byte[0])                                           ' lsb of address
  ackbit := write(value)                                        ' write one byte
  stop

  return ackbit


PUB putwordx(id, addr, value) | ackbit

'' Write word to device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 16-bit address in the device

  ackbit := putpagex(id, addr, 2, @value)

  return ackbit


PUB putlongx(id, addr, value) | ackbit

'' Write long to device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 16-bit address in the device

  ackbit := putpagex(id, addr, 4, @value)

  return ackbit  


PUB putpagex(id, addr, n, src) | ackbit

'' Write n bytes from src to device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 16-bit address in the device
'' --    n : bytes to write
'' --  src : pointer to source valuee
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr.byte[1])                                           ' msb of address
  write(addr.byte[0])                                           ' lsb of address
  ackbit := ACK                                                 ' assume okay
  repeat n
    ackbit |= write(byte[src++])                                ' write a byte 
  stop

  return ackbit


PUB getbytex(id, addr) | value

'' Read byte from device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  wait(id)
  write(addr.byte[1])                                           ' msb of address
  write(addr.byte[0])                                           ' lsb of address
  start                                                         ' restart for read
  write(id | $01)                                               ' read  
  value := read(NAK)                                            '  one byte
  stop

  return value & $FF


PUB getwordx(id, addr) | value

'' Read word from device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  getpagex(id, addr, 2, @value)

  return value & $FFFF
  

PUB getlongx(id, addr) | value

'' Read long from device w/16-bit addressing
'' --   id : device slave id
'' -- addr : 8-bit address in the device

  getpagex(id, addr, 4, @value)

  return value


PUB getpagex(id, addr, n, dest)

'' Read n bytes from device w/16-bit addressing; output to dest
'' --   id : device slave id (e.g., $A0 for EEPROM)
'' -- addr : 16-bit address in the device
'' --    n : bytes to read
'' -- dest : pointer to the destination
''
'' Be mindful of address/page size in device to prevent page wrap-around

  wait(id)
  write(addr.byte[1])                                           ' msb of address
  write(addr.byte[0])                                           ' lsb of address
  start                                                         ' restart for read
  write(id | $01)                                               ' device read
  repeat while n > 1
    byte[dest++] := read(ACK)
    --n
  byte[dest] := read(NAK)                                       ' last byte gets NAK
  stop 

 
PUB wait(id) | ackbit

'' Waits for I2C device to be ready for new command

  repeat
    start
    ackbit := write(id & $FE)
  until ackbit == ACK


PUB start

'' Create I2C start sequence

  outa[scl]~~                                                   ' SCL HIGH
  dira[scl]~~
  dira[sda]~                                                    ' SDA high (float to p/u)
  outa[sda]~
  dira[sda]~~                                                   ' SDA low 


PUB write(i2cbyte) | ackbit

'' Write byte to I2C buss

  outa[scl]~                                                    ' SCL low 
  dira[scl]~~                                                   

  i2cbyte <<= (32-8)                                            ' move msb (bit7) to bit31
  repeat 8                                                      ' output eight bits
    dira[sda] := ((i2cbyte <-= 1) ^ 1)                          ' send msb first
    outa[scl]~~                                                 ' SCL high
    outa[scl]~                                                  ' SCL low

  dira[sda]~                                                    ' relase to read ack bit
  outa[scl]~~                                                   ' SCL high 
  ackbit := ina[sda]                                            ' read ack bit
  outa[scl]~                                                    ' SCL low

  return ackbit


PUB read(ackbit) | i2cbyte

'' Read byte from I2C buss

  outa[scl]~                                                    ' prep to write low
  dira[scl]~~
  dira[sda]~                                                    ' make input for read

  repeat 8
    outa[scl]~~                                                 ' SCL high (float to p/u)
    i2cbyte := (i2cbyte << 1) | ina[sda]                        ' read the bit
    outa[scl]~                                                  ' SCL low
                             
  dira[sda] := !ackbit                                          ' output ack bit 
  outa[scl]~~                                                   ' clock it
  outa[scl]~

  return i2cbyte & $FF


PUB stop

'' Create I2C stop sequence 

  outa[sda]~
  dira[sda]~~                                                   ' SDA low
  outa[scl]~~
  dira[scl]~~
  dira[sda]~                                                    ' float SDA


DAT

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}